@@ -17,9 +17,10 @@ from websockets.server import WebSocketServerProtocol
 from websockets.http import Headers
 from websockets.http11 import Response
 
-# Suppress noisy websockets handshake errors (TLS mismatch, etc.)
-logging.getLogger("websockets.server").setLevel(logging.ERROR)
-logging.getLogger("websockets.protocol").setLevel(logging.ERROR)
+# Suppress noisy websockets handshake errors (TLS mismatch, InvalidUpgrade, etc.)
+logging.getLogger("websockets.server").setLevel(logging.CRITICAL)
+logging.getLogger("websockets.protocol").setLevel(logging.CRITICAL)
+logging.getLogger("websockets").setLevel(logging.CRITICAL)
 
 from darkcode_server.config import ServerConfig
 from darkcode_server.security import (
@@ -297,8 +298,27 @@ class DarkCodeServer:
             path = str(connection) if isinstance(connection, str) else getattr(connection, 'path', '/')
             request_headers = request if hasattr(request, 'get') else {}
 
+        # Serve favicon.ico
+        if isinstance(path, str) and path == '/favicon.ico':
+            try:
+                from darkcode_server.web_admin import serve_favicon
+                status, resp_headers, resp_body = serve_favicon()
+                header_list = [(k, v) for k, v in resp_headers.items()]
+                return Response(status, "OK", Headers(header_list), resp_body)
+            except ImportError:
+                return Response(404, "Not Found", Headers([]), b"")
+
+        # Check if web admin is disabled
+        if getattr(self.config, 'web_admin_disabled', False):
+            # Return None to let websockets handle it normally
+            # This may result in 426 for non-WS requests, which is fine
+            return None
+
         # Handle admin dashboard requests (non-WebSocket HTTP)
         if isinstance(path, str) and path.startswith('/admin'):
+            # Debug: print the full path to terminal
+            print(f"[DEBUG] Admin request path: {path!r}")
+
             # Initialize web admin handler if needed
             if self._web_admin is None:
                 try:
@@ -323,9 +343,11 @@ class DarkCodeServer:
             body = b''
 
             # Handle the request
+            print(f"[DEBUG] Calling handle_request with path={path!r}, method={method}")
             status, resp_headers, resp_body = self._web_admin.handle_request(
                 path, method, headers, body
             )
+            print(f"[DEBUG] Response: status={status}, headers={resp_headers}")
 
             # Build Response object for websockets 13+
             # Get status phrase
@@ -337,10 +359,37 @@ class DarkCodeServer:
 
             # Convert headers to Headers object
             header_list = [(k, v) for k, v in resp_headers.items()]
+            print(f"[DEBUG] Header list: {header_list}")
             response_headers = Headers(header_list)
 
             return Response(status, reason, response_headers, resp_body)
 
+        # Check if this is a WebSocket upgrade request
+        # If not, return a friendly HTTP response instead of letting websockets raise InvalidUpgrade
+        connection_header = request_headers.get('Connection', '').lower() if hasattr(request_headers, 'get') else ''
+        upgrade_header = request_headers.get('Upgrade', '').lower() if hasattr(request_headers, 'get') else ''
+
+        if 'upgrade' not in connection_header or upgrade_header != 'websocket':
+            # Non-WebSocket HTTP request - return a simple response
+            # This prevents InvalidUpgrade errors from browser auto-refresh, health checks, etc.
+            html = b"""<!DOCTYPE html>
+<html><head><title>DarkCode Server</title></head>
+<body style="font-family: system-ui; background: #1a1a2e; color: #eee; display: flex; justify-content: center; align-items: center; height: 100vh; margin: 0;">
+<div style="text-align: center;">
+<h1 style="color: #00d4ff;">DarkCode Server</h1>
+<p>WebSocket server running. Use the DarkCode app to connect.</p>
+<p><a href="/admin" style="color: #00d4ff;">Web Admin Dashboard</a></p>
+</div></body></html>"""
+            return Response(
+                200, "OK",
+                Headers([
+                    ("Content-Type", "text/html; charset=utf-8"),
+                    ("Content-Length", str(len(html))),
+                    ("Cache-Control", "no-cache"),
+                ]),
+                html
+            )
+
         # For WebSocket paths, return None to continue with upgrade
         return None
 
@@ -607,6 +656,7 @@ class DarkCodeServer:
                             "success": True,
                             "sessionId": session_id,
                             "workingDir": str(self.config.working_dir),
+                            "browseDir": str(self.config.effective_browse_dir),
                             "deviceBound": self.config.device_lock,
                             "state": self._state.value,
                             "isGuest": is_guest,
@@ -1187,6 +1237,103 @@ class DarkCodeServer:
             path = msg.get("path", "")
             await self._handle_delete_file(session, path)
 
+        # === Terminal/Bash Execution ===
+        elif msg_type == "execute_bash":
+            command = msg.get("command", "")
+            timeout = msg.get("timeout", 30000)  # Default 30s
+            await self._handle_execute_bash(session, command, timeout)
+
+    async def _handle_execute_bash(self, session: Session, command: str, timeout_ms: int):
+        """Execute a bash command directly and return the output.
+
+        SECURITY: Commands are executed in the session's working directory.
+        This is for terminal mode - direct shell access for the user.
+        """
+        if not command or not command.strip():
+            await session.websocket.send(json.dumps({
+                "type": "bash_output",
+                "command": command,
+                "output": "Error: Empty command",
+                "exitCode": 1,
+                "isError": True,
+            }))
+            return
+
+        # Sanitize command - basic safety checks
+        command = command.strip()
+
+        # Max command length
+        if len(command) > 10000:
+            await session.websocket.send(json.dumps({
+                "type": "bash_output",
+                "command": command[:100] + "...",
+                "output": "Error: Command too long",
+                "exitCode": 1,
+                "isError": True,
+            }))
+            return
+
+        timeout_sec = min(timeout_ms / 1000, 300)  # Max 5 minutes
+
+        try:
+            import asyncio
+
+            # Run command in working directory
+            process = await asyncio.create_subprocess_shell(
+                command,
+                stdout=asyncio.subprocess.PIPE,
+                stderr=asyncio.subprocess.PIPE,
+                cwd=str(session.working_dir),
+                env={**os.environ, "TERM": "xterm-256color"},
+            )
+
+            try:
+                stdout, stderr = await asyncio.wait_for(
+                    process.communicate(),
+                    timeout=timeout_sec
+                )
+                exit_code = process.returncode or 0
+
+                # Combine stdout and stderr
+                output = ""
+                if stdout:
+                    output += stdout.decode("utf-8", errors="replace")
+                if stderr:
+                    if output:
+                        output += "\n"
+                    output += stderr.decode("utf-8", errors="replace")
+
+                await session.websocket.send(json.dumps({
+                    "type": "bash_output",
+                    "command": command,
+                    "output": output.strip() or "(no output)",
+                    "exitCode": exit_code,
+                    "isError": exit_code != 0,
+                    "timestamp": int(time.time() * 1000),
+                }))
+
+            except asyncio.TimeoutError:
+                process.kill()
+                await session.websocket.send(json.dumps({
+                    "type": "bash_output",
+                    "command": command,
+                    "output": f"Error: Command timed out after {timeout_sec}s",
+                    "exitCode": 124,  # Standard timeout exit code
+                    "isError": True,
+                    "timestamp": int(time.time() * 1000),
+                }))
+
+        except Exception as e:
+            logger.error(f"Bash execution error: {e}")
+            await session.websocket.send(json.dumps({
+                "type": "bash_output",
+                "command": command,
+                "output": f"Error: {str(e)}",
+                "exitCode": 1,
+                "isError": True,
+                "timestamp": int(time.time() * 1000),
+            }))
+
     def _resolve_safe_path(self, session: Session, path: str) -> Optional[Path]:
         """Resolve a path safely within the working directory.
 
